import 'dart:convert';
import 'package:dartz/dartz.dart';
import 'orm_utills.dart';
import 'orm_domain_types.dart';
import 'package:intl/intl.dart';
import 'package:data_db/data_db.dart';

class Event {
  Function()? action;
  dynamic execute() => action!();
}

class Event_executeCommand {
  Future<Either<ErrorSqlResult, ExecuteSuccesSqlResult>> Function(
      String methodName, String sql)? action;
  Future<Either<ErrorSqlResult, ExecuteSuccesSqlResult>> execute(
          String methodName, String sql) =>
      action!(methodName, sql);
}

class Event_openQuery {
  Future<Either<ErrorSqlResult, SelectSuccesSqlResult>> Function(
      String methodName, String sql)? action;
  Future<Either<ErrorSqlResult, SelectSuccesSqlResult>> execute(
          String methodName, String sql) =>
      action!(methodName, sql);
}

class DatasetEvent {
  Function(Dataset dataset)? action;
  dynamic execute(Dataset dataset) => action!(dataset);
}

abstract class GenericField {
  GenericDataFields? _owner;
  final String _fieldName;
  final bool _notNullable;
  final bool _isPrimaryKey;
  bool _assigned = false;
  bool _modefied = false;
  String? _internalProblem;
  // ignore: unused_field
  final Type _dbType = DynamicDataField;

  // bool get notNullable => _notNullable;
  // bool get assigned => _assigned;
  // bool get modefied => _modefied;
  // String get fieldName => _fieldName;
  String get columnNameToSelect => _fieldName;

  GenericField(
      this._owner, this._fieldName, this._notNullable, this._isPrimaryKey) {
    //_owner._fields.add(this);
    //_owner._currentRecord.add(this);
  }

  GenericField.clone(GenericField field)
      : _fieldName = field._fieldName,
        _notNullable = field._notNullable,
        _isPrimaryKey = field._isPrimaryKey,
        _internalProblem = field._internalProblem;

  String _getSqlValue() => '';

  void _verify() {}
  void clear() {
    _assigned = false;
    _modefied = false;
  }

  void copy(GenericField field) {
    _internalProblem = field._internalProblem;
    _assigned = field._assigned;
    _modefied = field._modefied;
  }

  bool _trySetValue(dynamic value);
  bool _tryPutValue(void Function(dynamic value) fn);

  bool loadDataFromDataset(Dataset dataset) {
    return dataset
        .fieldByName(_fieldName)
        .fold(() => false, (a) => a.tryGetValue(_trySetValue));
  }

  bool setDateToDataset(Dataset dataset) {
    return dataset.fieldByName(_fieldName).fold(() => false, (a) {
      return _tryPutValue(a.setValue);
    });
  }

  String toJson() => '{$_fieldName: null}';
}

abstract class IntegerField extends GenericField {
  int _value = 0;
  @override
  // ignore: unused_field
  final Type _dbType = IntegerDataField;

  IntegerField(GenericDataFields owner, String fieldName, bool notNullable,
      bool isPrimaryKey)
      : super(owner, fieldName, notNullable, isPrimaryKey);

  IntegerField.clone(IntegerField field) : super.clone(field) {
    setValue(field._value);
  }

  Option<int> getValue() => option(_assigned, _value);

  int setValue(int value) {
    _owner?._change();
    _value = value;
    _assigned = true;
    _modefied = true;
    return value;
  }

  @override
  void copy(GenericField field) {
    super.copy(field);
    if (field is IntegerField) {
      _value = field._value;
    }
  }

  @override
  bool _trySetValue(dynamic value) {
    if (value is int) {
      return setValue(value) == _value;
    } else {
      return false;
    }
  }

  //set value(int v) => setValue(v);

  String toStringQuoted() => quotedStr(_value.toString());

  @override
  String _getSqlValue() => toStringQuoted();

  @override
  void clear() {
    super.clear();
    _value = 0;
  }

  @override
  bool _tryPutValue(void Function(dynamic value) fn) {
    return getValue().fold(() => false, (a) {
      fn(a);
      return true;
    });
  }

  @override
  String toJson() {
    return getValue().fold(() {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(null)}';
    }, (a) {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(a)}';
    });
  }
}

abstract class IntegerField_WithDefault extends IntegerField {
  final int _defaultValue;

  int get defaultValue => _defaultValue;

  IntegerField_WithDefault(GenericDataFields owner, String fieldName,
      bool notNullable, bool isPrimaryKey, this._defaultValue)
      : super(owner, fieldName, notNullable, isPrimaryKey) {
    _value = _defaultValue;
    _assigned = true;
  }

  IntegerField_WithDefault.clone(IntegerField_WithDefault field)
      : _defaultValue = field._defaultValue,
        super.clone(field) {
    setValue(field._value);
  }

  @override
  void clear() {
    super.clear();
    _value = _defaultValue;
  }
}

abstract class NumericField extends GenericField {
  @override
  // ignore: unused_field
  final Type _dbType = NumericDataField;
  double _value = 0;

  NumericField(GenericDataFields owner, String fieldName, bool notNullable,
      bool isPrimaryKey)
      : super(owner, fieldName, notNullable, isPrimaryKey);

  NumericField.clone(NumericField field) : super.clone(field) {
    setValue(field._value);
  }

  Option<double> getValue() => option(_assigned, _value);

  double setValue(double value) {
    _owner?._change();
    _value = value;
    _assigned = true;
    _modefied = true;
    return value;
  }

  @override
  void copy(GenericField field) {
    super.copy(field);
    if (field is NumericField) {
      _value = field._value;
    }
  }

  String toStringQuoted() => quotedStr(_value.toString());

  @override
  String _getSqlValue() => toStringQuoted();

  @override
  void clear() {
    super.clear();
    _value = _value;
  }

  @override
  bool _trySetValue(dynamic value) {
    if (value is double) {
      return setValue(value) == _value;
    } else {
      return false;
    }
  }

  @override
  bool _tryPutValue(void Function(dynamic value) fn) {
    return getValue().fold(() => false, (a) {
      fn(a);
      return true;
    });
  }

  @override
  String toJson() {
    return getValue().fold(() {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(null)}';
    }, (a) {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(a)}';
    });
  }
}

abstract class NumericField_WithDefault extends NumericField {
  final double _defaultValue;

  NumericField_WithDefault(GenericDataFields owner, String fieldName,
      bool notNullable, bool isPrimaryKey, this._defaultValue)
      : super(owner, fieldName, notNullable, isPrimaryKey) {
    _value = _defaultValue;
    _assigned = true;
  }

  NumericField_WithDefault.clone(NumericField_WithDefault field)
      : _defaultValue = field._defaultValue,
        super.clone(field) {
    setValue(field._value);
  }

  @override
  void clear() {
    super.clear();
    _value = _defaultValue;
  }
}

abstract class StringField extends GenericField {
  @override
  // ignore: unused_field
  final Type _dbType = StringDataField;
  String _value = '';
  final int _maxLength;

  StringField(GenericDataFields owner, String fieldName, bool notNullable,
      bool isPrimaryKey, this._maxLength)
      : super(owner, fieldName, notNullable, isPrimaryKey);

  StringField.clone(StringField field)
      : _maxLength = field._maxLength,
        super.clone(field) {
    setValue(field._value);
  }

  Option<String> getValue() => option(
      (_assigned && (_maxLength == -1 || (_value.length <= _maxLength))),
      _value);

  String setValue(String value) {
    _owner?._change();
    _value = value;
    _assigned = true;
    _modefied = true;
    if ((_maxLength > 0) && (_value.length > _maxLength)) {
      _internalProblem =
          'O conteúdo campo ${quotedStr(_fieldName)} está maior do que o tamanho permitido($_maxLength).';
    }
    return value;
  }

  @override
  void copy(GenericField field) {
    super.copy(field);
    if (field is StringField) {
      _value = field._value;
    }
  }

  String toStringQuoted() => quotedStr(_value);

  @override
  String _getSqlValue() => toStringQuoted();

  @override
  void clear() {
    super.clear();
    _value = _value;
  }

  @override
  bool _trySetValue(dynamic value) {
    if (value is String) {
      return setValue(value) == _value;
    } else {
      return false;
    }
  }

  @override
  bool _tryPutValue(void Function(dynamic value) fn) {
    return getValue().fold(() => false, (a) {
      fn(a);
      return true;
    });
  }

  @override
  String toJson() {
    return getValue().fold(() {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(null)}';
    }, (a) {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(a)}';
    });
  }
}

abstract class StringField_WithDefault extends StringField {
  final String _defaultValue;

  StringField_WithDefault(GenericDataFields owner, String fieldName,
      bool notNullable, bool isPrimaryKey, maxLength, this._defaultValue)
      : super(owner, fieldName, notNullable, isPrimaryKey, maxLength) {
    _value = _defaultValue;
    _assigned = true;
  }

  StringField_WithDefault.clone(StringField_WithDefault field)
      : _defaultValue = field._defaultValue,
        super.clone(field) {
    setValue(field._value);
  }

  @override
  void clear() {
    super.clear();
    _value = _defaultValue;
  }
}

abstract class BooleanField extends GenericField {
  @override
  // ignore: unused_field
  final Type _dbType = BooleanDataField;
  bool _value = false;

  BooleanField(GenericDataFields owner, String fieldName, bool notNullable,
      bool isPrimaryKey)
      : super(owner, fieldName, notNullable, isPrimaryKey);

  BooleanField.clone(BooleanField field) : super.clone(field) {
    setValue(field._value);
  }

  Option<bool> getValue() => option(_assigned, _value);

  bool setValue(bool value) {
    _owner?._change();
    _value = value;
    _assigned = true;
    _modefied = true;
    return value;
  }

  @override
  void copy(GenericField field) {
    super.copy(field);
    if (field is BooleanField) {
      _value = field._value;
    }
  }

  String toStringQuoted() => quotedStr(_value.toString());

  @override
  String _getSqlValue() => _value.toString();

  @override
  void clear() {
    super.clear();
    _value = _value;
  }

  @override
  bool _trySetValue(dynamic value) {
    if (value is bool) {
      return setValue(value) == _value;
    } else {
      return false;
    }
  }

  @override
  bool _tryPutValue(void Function(dynamic value) fn) {
    return getValue().fold(() => false, (a) {
      fn(a);
      return true;
    });
  }

  @override
  String toJson() {
    return getValue().fold(() {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(null)}';
    }, (a) {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(a)}';
    });
  }
}

abstract class BooleanField_WithDefault extends BooleanField {
  final bool _defaultValue;

  BooleanField_WithDefault(GenericDataFields owner, String fieldName,
      bool notNullable, bool isPrimaryKey, this._defaultValue)
      : super(owner, fieldName, notNullable, isPrimaryKey) {
    _value = _defaultValue;
    _assigned = true;
  }

  BooleanField_WithDefault.clone(BooleanField_WithDefault field)
      : _defaultValue = field._defaultValue,
        super.clone(field) {
    setValue(field._value);
  }

  @override
  void clear() {
    super.clear();
    _value = _defaultValue;
  }
}

abstract class DateField extends GenericField {
  @override
  // ignore: unused_field
  final Type _dbType = DateDataField;
  DateTime _value = DateTime(2020);

  DateField(GenericDataFields owner, String fieldName, bool notNullable,
      bool isPrimaryKey)
      : super(owner, fieldName, notNullable, isPrimaryKey);

  DateField.clone(DateField field) : super.clone(field) {
    setValue(field._value);
  }

  Option<DateTime> getValue() => option(_assigned, _value);

  DateTime setValue(DateTime value) {
    _owner?._change();
    DateTime.utc(_value.year, _value.month, _value.day);
    _assigned = true;
    _modefied = true;
    return _value;
  }

  @override
  void copy(GenericField field) {
    super.copy(field);
    if (field is DateField) {
      _value = field._value;
    }
  }

  String toStringQuoted() => quotedStr(format('yyyy-MM-dd'));

  String format(String format) {
    return DateFormat(format).format(_value);
  }

  @override
  String _getSqlValue() => toStringQuoted();

  @override
  void clear() {
    super.clear();
    _value = _value;
  }

  @override
  bool _trySetValue(dynamic value) {
    if (value is DateTime) {
      return setValue(value) == _value;
    } else {
      return false;
    }
  }

  @override
  bool _tryPutValue(void Function(dynamic value) fn) {
    return getValue().fold(() => false, (a) {
      fn(a);
      return true;
    });
  }

  @override
  String toJson() {
    return getValue().fold(() {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(null)}';
    }, (a) {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(a.toString())}';
    });
  }
}

abstract class DateField_WithDefault extends DateField {
  late final DateTime _defaultValue;

  DateField_WithDefault(GenericDataFields owner, String fieldName,
      bool notNullable, bool isPrimaryKey, DateTime defaultValue)
      : super(owner, fieldName, notNullable, isPrimaryKey) {
    _defaultValue =
        DateTime(defaultValue.year, defaultValue.month, defaultValue.day);
    _value = _defaultValue;
    _assigned = true;
  }

  DateField_WithDefault.clone(DateField_WithDefault field)
      : _defaultValue = field._defaultValue,
        super.clone(field) {
    setValue(field._value);
  }
}

abstract class DateTimeField extends GenericField {
  @override
  // ignore: unused_field
  final Type _dbType = DateTimeDataField;
  DateTime _value = DateTime(2020);

  DateTimeField(GenericDataFields owner, String fieldName, bool notNullable,
      bool isPrimaryKey)
      : super(owner, fieldName, notNullable, isPrimaryKey);

  DateTimeField.clone(DateTimeField field) : super.clone(field) {
    setValue(field._value);
  }

  Option<DateTime> getValue() => option(_assigned, _value);

  DateTime setValue(DateTime value) {
    _owner?._change();
    _value = value;
    _assigned = true;
    _modefied = true;
    return _value;
  }

  @override
  void copy(GenericField field) {
    super.copy(field);
    if (field is DateTimeField) {
      _value = field._value;
    }
  }

  String toStringQuoted() => quotedStr(format('yyyy-MM-dd HH:mm:ss'));

  String format(String format) {
    return DateFormat(format).format(_value);
  }

  @override
  String _getSqlValue() => toStringQuoted();

  @override
  bool _trySetValue(dynamic value) {
    if (value is DateTime) {
      return setValue(value) ==
          DateTime.utc(_value.year, _value.month, _value.day, _value.hour,
              _value.minute, _value.second);
    } else {
      return false;
    }
  }

  @override
  bool _tryPutValue(void Function(dynamic value) fn) {
    return getValue().fold(() => false, (a) {
      fn(a);
      return true;
    });
  }

  @override
  String toJson() {
    return getValue().fold(() {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(null)}';
    }, (a) {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(a.toString())}';
    });
  }
}

abstract class DateTimeField_WithDefault extends DateTimeField {
  final DateTime _defaultValue;

  DateTimeField_WithDefault(GenericDataFields owner, String fieldName,
      bool notNullable, bool isPrimaryKey, this._defaultValue)
      : super(owner, fieldName, notNullable, isPrimaryKey) {
    _value = _defaultValue;
    _assigned = true;
  }

  DateTimeField_WithDefault.clone(DateTimeField_WithDefault field)
      : _defaultValue = field._defaultValue,
        super.clone(field) {
    setValue(field._value);
  }
}

abstract class TimeField extends GenericField {
  @override
  // ignore: unused_field
  final Type _dbType = TimeDataField;
  DateTime _value = DateTime(2020);

  TimeField(GenericDataFields owner, String fieldName, bool notNullable,
      bool isPrimaryKey)
      : super(owner, fieldName, notNullable, isPrimaryKey);

  TimeField.clone(TimeField field) : super.clone(field) {
    setValue(field._value);
  }

  Option<DateTime> getValue() => option(_assigned, _value);

  DateTime setValue(DateTime value) {
    _owner?._change();
    _value = DateTime.utc(
        0, 0, 0, value.hour, value.minute, value.second, value.millisecond);
    _assigned = true;
    _modefied = true;
    return _value;
  }

  @override
  void copy(GenericField field) {
    super.copy(field);
    if (field is DateTimeField) {
      _value = field._value;
    }
  }

  String toStringQuoted() => quotedStr(format('HH:mm:ss'));

  String format(String format) {
    return DateFormat(format).format(_value);
  }

  @override
  String _getSqlValue() => toStringQuoted();

  @override
  bool _trySetValue(dynamic value) {
    if (value is DateTime) {
      return setValue(value) == _value;
    } else {
      return false;
    }
  }

  @override
  bool _tryPutValue(void Function(dynamic value) fn) {
    return getValue().fold(() => false, (a) {
      fn(a);
      return true;
    });
  }

  @override
  String toJson() {
    return getValue().fold(() {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(null)}';
    }, (a) {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(a.toString())}';
    });
  }
}

abstract class TimeField_WithDefault extends TimeField {
  final DateTime _defaultValue;

  TimeField_WithDefault(GenericDataFields owner, String fieldName,
      bool notNullable, bool isPrimaryKey, this._defaultValue)
      : super(owner, fieldName, notNullable, isPrimaryKey) {
    _value = _defaultValue;
    _assigned = true;
  }

  TimeField_WithDefault.clone(TimeField_WithDefault field)
      : _defaultValue = field._defaultValue,
        super.clone(field) {
    setValue(field._value);
  }
}

enum DataFieldStatus { emptyMode, readyMode, insertMode, editMode }

GenericDataFields? getBackup(GenericDataFields origin) => origin._backup;
void setBackup(GenericDataFields origin, GenericDataFields backup) =>
    origin._backup = backup;

abstract class GenericDataFields {
  late GenericDataFields _backup;
  DataFieldStatus status = DataFieldStatus.emptyMode;
  late bool _assigned = false;
  late List<GenericField> fields;

  bool get isReady => status == DataFieldStatus.readyMode;
  bool get isNotready => !isReady;
  bool get assigned => _assigned;

  bool get inChange => ((status == DataFieldStatus.editMode) ||
      (status == DataFieldStatus.insertMode));
  bool get notInChange => !inChange;

  void _change() {
    if (notInChange) {
      if (_assigned) {
        try {
          backup();
          _backup._assigned = true;
          _backup.status = DataFieldStatus.readyMode;
          status = DataFieldStatus.editMode;
        } on Exception catch (e) {
          print(e.toString());
        }
      } else {
        status = DataFieldStatus.insertMode;
      }
    }
  }

  void backup() {}
  void restore() {}

  bool contains(List<GenericField> fields) {
    var x = 0;
    int y;

    fields.forEach((externalField) {
      y = (this.fields.where((field) {
        return (field._fieldName == externalField._fieldName) &&
            (field._assigned == externalField._assigned) &&
            (field._getSqlValue() == externalField._getSqlValue());
      })).length;
      if (y == 1) {
        x++;
      }
    });

    if (x == fields.length) {
      return true;
    } else {
      return false;
    }
  }

  void finalize() {
    _assigned = true;
    status = DataFieldStatus.readyMode;
  }
}

abstract class ORM {
  final SqlConnection _sqlConnection;
  final String _schemaName;
  final String _tableName;
  final List<GenericDataFields> _records = <GenericDataFields>[];
  int _rowIndex = 0;
  late final String _className =
      toString().replaceFirst('Instance of ', '').replaceAll("'", '');

  //events
  final Event _addRecord;

  int get rowIndex => _rowIndex;
  int get count => _records.where((element) => element.assigned).length;

  GenericDataFields get _currentRecord => _records[_rowIndex];

  String get _fullTableName => '$_schemaName.$_tableName';

  ORM(this._sqlConnection, this._schemaName, this._tableName, this._addRecord);

  Future<Either<ErrorSqlResult, ExecuteSuccesSqlResult>> insert() async {
    return (_prepareInsertSQL()).fold((l) => left(l..methodName = 'insert'),
        (r) async {
      return (await _sqlConnection.executeCommand('insert', r))
          .fold((l) => left(l), (r) {
        if (_currentRecord.assigned) {
          var x = _records.removeAt(_rowIndex);
          _records.add(x._backup);
          _records.add(x);
          _rowIndex++;
        }
        _currentRecord.finalize();

        return right(r);
      });
    });
  }

  Future<Either<ErrorSqlResult, ExecuteSuccesSqlResult>> update() async {
    return (_prepareUpdateSQL()).fold((l) => left(l..methodName = 'update'),
        (sql) async {
      return (await _sqlConnection.executeCommand('update', sql))
          .fold((l) => left(l), (r) {
        if (r.rowsAffected == 1) {
          _currentRecord.finalize();
          return right(r);
        } else {
          return left(
              ErrorSqlResult(_className, 'update', sql, ['record not found']));
        }
      });
    });
  }

  Future<Either<ErrorSqlResult, ExecuteSuccesSqlResult>> delete() async {
    return (_prepareDeleteSQL()).fold((l) => left(l..methodName = 'delete'),
        (sql) async {
      return (await _sqlConnection.executeCommand('delete', sql))
          .fold((l) => left(l), (r) {
        if (r.rowsAffected == 1) {
          _records.removeAt(_rowIndex);
          if (_rowIndex == 0) {
            _addRecord.execute();
          } else {
            _rowIndex--;
          }
          return right(r);
        } else {
          return left(
              ErrorSqlResult(_className, 'delete', sql, ['record not found']));
        }
      });
    });
  }

  Future<Either<ErrorSqlResult, ExecuteSuccesSqlResult>> _deleteRecord(
      List<GenericField> fields) async {
    return (_prepareDeleteREcordSQL(fields))
        .fold((l) => left(l..methodName = 'deleteRecord'), (sql) async {
      return (await _sqlConnection.executeCommand('deleteRecord', sql))
          .fold((l) => left(l), (r) {
        if (r.rowsAffected >= 1) {
          cancelAllChanges();
          var x = locateField(fields);
          if (x > -1) {
            _records.removeAt(x);
            if (_rowIndex > 0) {
              _rowIndex--;
            } else {
              _addRecord.execute();
            }
          }
          return right(r);
        } else {
          return left(ErrorSqlResult(
              _className, 'deleteRecord', sql, ['record not found']));
        }
      });
    });
  }

  Future<Either<ErrorSqlResult, SelectSuccesSqlResult>> _materialize(
      List<GenericField> fields) async {
    if (locate(fields)) {
      return right(SelectSuccesSqlResult(_className, 'materialize', 1, []));
    } else {
      return (_prepareMaterializeSQL(fields)).fold((l) {
        return left(l..methodName = 'materialize');
      }, (sql) async {
        return (await _sqlConnection.openQuery('materialize', sql)).fold((l) {
          return left(l);
        }, (res) {
          if (res.dataset.count > 0) {
            if (count > 0) {
              _addRecord.execute();
              _rowIndex++;
            }

            return loadFromDataset(res.dataset).fold((l) {
              return left(ErrorSqlResult(_className, 'materialize', '', [l]));
            }, (r) {
              return right(res);
            });
          } else {
            return left(ErrorSqlResult(
                _className, 'materialize', sql, ['record not found']));
          }
        });
      });
    }
  }

  Either<ErrorSqlResult, String> _prepareDeleteREcordSQL(
      List<GenericField> fields) {
    return getSQL(getDeleteSQL(fields, _fullTableName));
  }

  Either<ErrorSqlResult, String> _prepareMaterializeSQL(
      List<GenericField> fields) {
    return getSQL(getMaterializeSQL(fields, _fullTableName,
        columns: getColumnsToSelect(_currentRecord.fields)));
  }

  Either<ErrorSqlResult, String> _prepareInsertSQL() {
    return verifyAllFields(_currentRecord.fields).fold(
        (problemFields) => left(ErrorSqlResult(
            _className, '_prepareInsertSQL', null, problemFields)),
        (assignedFields) =>
            getSQL(getInsertSQL(assignedFields, _fullTableName)));
  }

  Either<ErrorSqlResult, String> _prepareUpdateSQL() {
    return verifyAllFields(_currentRecord.fields).fold(
        (problemFields) => left(ErrorSqlResult(
            _className, '_prepareUpdateSQL', null, problemFields)),
        (assignedFields) =>
            getSQL(getUpdateSQL(assignedFields, _fullTableName)));
  }

  Either<ErrorSqlResult, String> _prepareDeleteSQL() {
    return verifyPKFields(_currentRecord.fields).fold(
        (problemFields) => left(ErrorSqlResult(
            _className, '_prepareDeleteSQL', null, problemFields)),
        (assignedFields) =>
            getSQL(getDeleteSQL(assignedFields, _fullTableName)));
  }

  Either<String, int> loadFromDataset(Dataset dataset) {
    if (dataset.isNotempty) {
      dataset.first();
      _currentRecord.fields.forEach((element) {
        element.loadDataFromDataset(dataset);
      });

      _currentRecord.finalize();

      return right(dataset.count);
    } else {
      return left('Dataset is empty.');
    }
  }

  void cancelChanges() {
    if ((_currentRecord.inChange) & (_currentRecord.assigned)) {
      _currentRecord.restore();
    }
  }

  void cancelAllChanges() {
    _records.forEach((element) {
      if ((element.inChange) & (element.assigned)) {
        element.restore();
      }
    });
  }

  void clear() {
    _records.clear();
    _addRecord.execute();
    _rowIndex = 0;
  }

  bool locate(List<GenericField> fields) {
    var x = locateField(fields);
    if (x >= 0) {
      _rowIndex = x;
      return true;
    } else {
      return false;
    }
  }

  int locateField(List<GenericField> fields) {
    var x = _records.indexWhere((element) => element.contains(fields));
    if (x >= 0) {
      if (_records[x]._assigned) {
        if (_records[x].inChange) {
          _records[x].restore();
        }

        return x;
      } else {
        return -1;
      }
    } else {
      return -1;
    }
  }

  String toJson() {
    var x = '';
    var r = '';
    cancelAllChanges();
    _records.where((element) => element.isReady).forEach((element) {
      r = '';
      element.fields.forEach((field) {
        r += ', ${field.toJson()}';
      });
      x += ', {${r.replaceFirst(', ', '')}}';
    });

    return '[${x.replaceFirst(', ', '')}]';
  }
}

Either<String, List<GenericField>> _getPKFields(List<GenericField> fields) {
  var pk = fields.where((field) => field._isPrimaryKey);
  if (pk.isNotEmpty) {
    return right(pk.toList());
  } else {
    return left('No primary Key assigned.');
  }
}

Either<String, List<GenericField>> _getModefiedFields(
    List<GenericField> assignedFields) {
  var updatableFields =
      assignedFields.where((field) => !field._isPrimaryKey && field._modefied);
  if (updatableFields.isNotEmpty) {
    return right(updatableFields.toList());
  } else {
    return left('No modefied fields');
  }
}

String _getUpdateSyntaxFromFields(List<GenericField> fields) {
  var columns = '';

  fields.forEach((field) {
    columns += ', ${field._fieldName} = ${field._getSqlValue()}';
  });

  return columns.replaceFirst(', ', '');
}

String _getWhereSyntaxFromFields(List<GenericField> fields) {
  var primaryKey = '';

  fields.forEach((field) {
    primaryKey += 'and ${field._fieldName} = ${field._getSqlValue()}';
  });

  return primaryKey.replaceFirst('and ', '');
}

Either<String, String> getInsertSQL(
    List<GenericField> assignedFields, String fullTableName) {
  var columns = '';
  var values = '';
  var sql = 'insert into $fullTableName';

  try {
    assignedFields.forEach((field) {
      columns += ',${field._fieldName}';
      values += ',${field._getSqlValue()}';
    });

    sql +=
        ' (${columns.replaceFirst(',', '')}) values(${values.replaceFirst(',', '')})';
    return right(sql);
  } on Exception catch (e) {
    return left('Unknown exception: $e');
  }
}

Either<List<String>, List<GenericField>> verifyFieldsInternal(
    List<GenericField> assignedFields) {
  assignedFields.forEach((field) {
    field._verify();
  });

  var problemFields = getFieldsWith_InternalProblems(assignedFields);

  if (problemFields.isNotEmpty) {
    return left(getMessageFromProblemFields(problemFields));
  } else {
    return right(assignedFields);
  }
}

Either<List<String>, List<GenericField>> verifyAllnotNullableFields(
    List<GenericField> fields) {
  var notNullables = getNotNullableFields(fields);

  if (notNullables.isNotEmpty) {
    var requireds = getNotAssignedFields(notNullables);
    if (requireds.isNotEmpty) {
      return left(getMessageForrequiredFields(requireds));
    } else {
      return right(getAssignedFields(fields));
    }
  } else {
    return right(fields);
  }
}

Either<List<String>, List<GenericField>> verifyAllFields(
    List<GenericField> fields) {
  return verifyAllnotNullableFields(fields)
      .fold((l) => left(l), (r) => verifyFieldsInternal(r));
}

Either<String, String> getDeleteSQL(
    List<GenericField> assignedFields, String fullTableName) {
  return right(
      'delete from $fullTableName where ${_getWhereSyntaxFromFields(assignedFields)}');
}

Either<String, String> getMaterializeSQL(
    List<GenericField> pkFields, String fullTableName,
    {String columns = '*'}) {
  return right(
      'select $columns from $fullTableName where ${_getWhereSyntaxFromFields(pkFields)}');
}

Either<List<String>, List<GenericField>> verifyPKFields(
    List<GenericField> fields) {
  return _getPKFields(fields)
      .fold((l) => left([l]), (r) => verifyAllPKFields(r));
}

Either<List<String>, List<GenericField>> verifyAllPKFields(
    List<GenericField> pkFields) {
  var requireds = getNotAssignedFields(pkFields);
  if (requireds.isNotEmpty) {
    return left(getMessageForrequiredFields(requireds));
  } else {
    return right(getAssignedFields(pkFields));
  }
}

Either<String, String> getUpdateSQL(
    List<GenericField> assignedFields, String tableName) {
  return _getPKFields(assignedFields).fold(
      (l) => left(l),
      (pk) => _getModefiedFields(assignedFields).fold((l) => left(l),
          (r) => right(_getMountedUpdateSql(pk, r, tableName))));
}

String _getMountedUpdateSql(
    List<GenericField> pk, List<GenericField> up, String tablename) {
  return 'update $tablename set ${_getUpdateSyntaxFromFields(up)}  where ${_getWhereSyntaxFromFields(pk)}';
}

String getColumnsToSelect(List<GenericField> fields) {
  var x = '';
  fields.forEach((field) {
    x += ', ${field.columnNameToSelect}';
  });

  return x.replaceFirst(', ', '');
}

List<GenericField> getNotNullableFields(List<GenericField> fields) {
  return fields.where((field) => field._notNullable).toList();
}

List<GenericField> getFieldsWith_InternalProblems(List<GenericField> fields) {
  return fields.where((field) => field._internalProblem != null).toList();
}

List<GenericField> getNotAssignedFields(List<GenericField> fields) {
  return fields.where((field) => !field._assigned).toList();
}

List<GenericField> getAssignedFields(List<GenericField> fields) {
  return fields.where((field) => field._assigned).toList();
}

List<String> getMessageForrequiredFields(List<GenericField> fields) {
  var x = [''];
  x.clear();

  fields.forEach((field) {
    x.add('O campo ${field._fieldName} é obrigatório');
  });

  return x;
}

List<String> getMessageFromProblemFields(List<GenericField> fields) {
  var x = [''];
  x.clear();

  fields.forEach((field) {
    x.add(field._internalProblem ?? '');
  });

  return x;
}

Either<ErrorSqlResult, String> getSQL(Either<String, String> sql) {
  return sql.fold(
      (l) => left(ErrorSqlResult('', 'getSQL', null, [l])), (r) => right(r));
}

Future<Either<ErrorSqlResult, ExecuteSuccesSqlResult>> getDeleteRecord(
    ORM orm, List<GenericField> fields) {
  return orm._deleteRecord(fields);
}

Future<Either<ErrorSqlResult, SelectSuccesSqlResult>> getMaterialize(
    ORM orm, List<GenericField> fields) {
  return orm._materialize(fields);
}

List<GenericDataFields> getRecords(ORM orm) => orm._records;

String getFieldName(GenericField field) => field._fieldName;
